"""Unittests for email structures."""
from unittest import TestCase
from unittest.mock import Mock, call, sentinel
from uuid import UUID

from typist import SecretaryClient
from typist.client import TemplateNotFound

from hermes.email import Email, EmailEnvelope


class EmailTest(TestCase):
    """Unittests for Email structure."""

    def test_init_empty(self):
        email = Email(sentinel.recipient, sentinel.subject_template, sentinel.body_template)
        self.assertEqual(email.recipient, sentinel.recipient)
        self.assertEqual(email.subject_template, sentinel.subject_template)
        self.assertEqual(email.body_template, sentinel.body_template)
        self.assertIsNone(email.body_template_html)
        self.assertIsNone(email.sender)
        self.assertEqual(email.extra_headers, {})
        self.assertEqual(email.context, {})
        self.assertEqual(email.attachments, [])
        self.assertIsNone(email.type)

    def test_init_values(self):
        email = Email(sentinel.recipient, sentinel.subject_template, sentinel.body_template,
                      body_template_html=sentinel.body_template_html, sender=sentinel.sender,
                      extra_headers={'header': 'content'}, context={}, attachments=['a', 'b'],
                      subject_template_uuid=UUID(int=1), body_template_uuid=UUID(int=2),
                      body_template_html_uuid=UUID(int=3), type=sentinel.type,
                      secretary_client=sentinel.secretary_client)
        self.assertEqual(email.recipient, sentinel.recipient)
        self.assertEqual(email.subject_template, sentinel.subject_template)
        self.assertEqual(email.body_template, sentinel.body_template)
        self.assertEqual(email.body_template_html, sentinel.body_template_html)
        self.assertEqual(email.subject_template_uuid, UUID(int=1))
        self.assertEqual(email.body_template_uuid, UUID(int=2))
        self.assertEqual(email.body_template_html_uuid, UUID(int=3))
        self.assertEqual(email.sender, sentinel.sender)
        self.assertEqual(email.extra_headers, {'header': 'content'})
        self.assertEqual(email.context, {})
        self.assertEqual(email.attachments, ['a', 'b'])
        self.assertEqual(email.secretary_client, sentinel.secretary_client)
        self.assertEqual(email.type, sentinel.type)

    def test_from_dict_missing_recipient(self):
        with self.assertRaises(ValueError):
            Email.from_dict({'subject_template': sentinel.subject_template,
                             'body_template': sentinel.body_template})

    def test_from_dict_missing_subject_template(self):
        with self.assertRaises(ValueError):
            Email.from_dict({'recipient': sentinel.recipient,
                             'body_template': sentinel.body_template})

    def test_from_dict_missing_body_template(self):
        with self.assertRaises(ValueError):
            Email.from_dict({'recipient': sentinel.recipient,
                             'subject_template': sentinel.subject_template})

    def test_from_dict_data(self):
        email = Email.from_dict(
            {'recipient': sentinel.recipient, 'subject_template': sentinel.subject_template,
             'body_template': sentinel.body_template, 'sender': sentinel.sender, 'context': {},
             'attachments': ['a', 'b'], 'extra_headers': {'header': 'content'},
             'subject_template_uuid': UUID(int=1), 'body_template_uuid': UUID(int=2), 'type': sentinel.type},
            secretary_client=sentinel.secretary_client,
        )
        self.assertEqual(email.recipient, sentinel.recipient)
        self.assertEqual(email.subject_template, sentinel.subject_template)
        self.assertEqual(email.body_template, sentinel.body_template)
        self.assertEqual(email.sender, sentinel.sender)
        self.assertEqual(email.extra_headers, {'header': 'content'})
        self.assertEqual(email.context, {})
        self.assertEqual(email.attachments, ['a', 'b'])
        self.assertEqual(email.subject_template_uuid, UUID(int=1))
        self.assertEqual(email.body_template_uuid, UUID(int=2))
        self.assertEqual(email.secretary_client, sentinel.secretary_client)
        self.assertEqual(email.type, sentinel.type)

    def test_render_without_secretary_client(self):
        email = Email(
            recipient='recipient@example.com',
            subject_template='subject_template',
            body_template='body_template',
            body_template_html='body_template_html',
        )
        with self.assertRaises(RuntimeError):
            email.subject
        with self.assertRaises(RuntimeError):
            email.body
        with self.assertRaises(RuntimeError):
            email.body_html

    def test_to_dict(self):
        email = Email.from_dict(
            {'recipient': sentinel.recipient, 'subject_template': sentinel.subject_template,
             'body_template': sentinel.body_template, 'sender': sentinel.sender, 'context': {},
             'attachments': ['a', 'b'], 'extra_headers': {'header': 'content'},
             'subject_template_uuid': UUID(int=1), 'body_template_uuid': UUID(int=2)},
        )
        self.assertEqual(email.to_dict(), {
            'sender': sentinel.sender,
            'recipient': sentinel.recipient,
            'extra_headers': {'header': 'content'},
            'subject_template': sentinel.subject_template,
            'body_template': sentinel.body_template,
            'body_template_html': None,
            'subject_template_uuid': UUID(int=1),
            'body_template_uuid': UUID(int=2),
            'body_template_html_uuid': None,
            'context': {},
            'attachments': ['a', 'b'],
        })


class EmailWithSecretaryClientTest(TestCase):
    """Unittests for Email structure with a secretary client."""

    def setUp(self):
        self.secretary_client = Mock(autospec=SecretaryClient)
        self.email = Email(
            recipient='recipient@example.com',
            subject_template='subject_template',
            body_template='body_template',
            body_template_html='body_template_html',
            context=sentinel.context,
            secretary_client=self.secretary_client,
        )

    def test_subject(self):
        self.secretary_client.render.return_value = sentinel.subject
        self.assertEqual(self.email.subject, sentinel.subject)
        self.assertEqual(self.email.subject, sentinel.subject)  # call subject again to test the caching
        self.assertEqual(
            self.secretary_client.mock_calls,
            [call.render('subject_template', context=sentinel.context)]
        )

    def test_body(self):
        self.secretary_client.render.return_value = sentinel.body
        self.assertEqual(self.email.body, sentinel.body)
        self.assertEqual(self.email.body, sentinel.body)  # call body again to test the caching
        self.assertEqual(
            self.secretary_client.mock_calls,
            [call.render('body_template', context=sentinel.context)]
        )

    def test_body_html(self):
        self.secretary_client.render_html.return_value = sentinel.body_html
        self.assertEqual(self.email.body_html, sentinel.body_html)
        self.assertEqual(self.email.body_html, sentinel.body_html)  # call body_html again to test the caching
        self.assertEqual(
            self.secretary_client.mock_calls,
            [call.render_html('body_template_html', context=sentinel.context)]
        )

    def test_body_html_not_set(self):
        email = Email('recipient@example.com', 'subject_template', 'body_template')
        self.assertIsNone(email.body_html)

    def test_to_dict(self):
        self.secretary_client.render.side_effect = [sentinel.subject, sentinel.body]
        self.secretary_client.render_html.return_value = sentinel.body_html
        self.assertEqual(self.email.to_dict(), {
            'sender': None,
            'recipient': 'recipient@example.com',
            'extra_headers': {},
            'subject': sentinel.subject,
            'body': sentinel.body,
            'body_html': sentinel.body_html,
            'subject_template': 'subject_template',
            'body_template': 'body_template',
            'body_template_html': 'body_template_html',
            'subject_template_uuid': None,
            'body_template_uuid': None,
            'body_template_html_uuid': None,
            'context': sentinel.context,
            'attachments': [],
        })

    def test_get_secretary_client_data_with_errors(self):
        for error_cls in [RuntimeError, TemplateNotFound, ValueError]:
            with self.subTest(error_cls=error_cls):
                self.secretary_client.render.side_effect = error_cls('Dangerous error')
                self.secretary_client.render_html.side_effect = error_cls('Dangerous error')
                with self.assertLogs('hermes.email', level='WARNING') as log:
                    data = self.email._get_secretary_client_data()
                self.assertEqual(data, {
                    'subject': None,
                    'body': None,
                    'body_html': None,
                    'render_errors': [
                        'SecretaryClient raised an error in the "subject" property: Dangerous error',
                        'SecretaryClient raised an error in the "body" property: Dangerous error',
                        'SecretaryClient raised an error in the "body_html" property: Dangerous error',
                    ],
                })
                self.assertEqual(log.output, [
                    'WARNING:hermes.email:SecretaryClient raised an error in the "subject" property: Dangerous error',
                    'WARNING:hermes.email:SecretaryClient raised an error in the "body" property: Dangerous error',
                    'WARNING:hermes.email:SecretaryClient raised an error in the "body_html" property: Dangerous error',
                ])

    def test_secretary_client_fields_error_raised(self):
        self.secretary_client.render.side_effect = TemplateNotFound()
        self.secretary_client.render_html.side_effect = TemplateNotFound()
        for item in ['subject', 'body', 'body_html'] * 2:  # call twice to test nothing has been cached
            with self.assertRaises(TemplateNotFound):
                getattr(self.email, item)


class EmailEnvelopeTest(TestCase):
    """Unittests for EmailEnvelope structure."""

    def test_init_empty(self):
        email = Email(sentinel.recipient, sentinel.subject_template, sentinel.body_template)
        email_envelope = EmailEnvelope(email)
        self.assertIsNone(email_envelope.uid)
        self.assertIsNone(email_envelope.create_datetime)
        self.assertIsNone(email_envelope.send_datetime)
        self.assertIsNone(email_envelope.delivery_datetime)
        self.assertEqual(email_envelope.attempts, 0)
        self.assertIsNone(email_envelope.status)
        self.assertEqual(email_envelope.message, email)
        self.assertTrue(email_envelope.archive)
        self.assertIsNone(email_envelope.message_id)

    def test_init_values(self):
        email = Email(sentinel.recipient, sentinel.subject_template, sentinel.body_template)
        email_envelope = EmailEnvelope(
            email, uid=sentinel.uid, create_datetime=sentinel.create_datetime, send_datetime=sentinel.send_datetime,
            delivery_datetime=sentinel.delivery_datetime, attempts=1, status=5, archive=False,
            message_id=sentinel.message_id)
        self.assertEqual(email_envelope.uid, sentinel.uid)
        self.assertEqual(email_envelope.create_datetime, sentinel.create_datetime)
        self.assertEqual(email_envelope.send_datetime, sentinel.send_datetime)
        self.assertEqual(email_envelope.delivery_datetime, sentinel.delivery_datetime)
        self.assertEqual(email_envelope.attempts, 1)
        self.assertEqual(email_envelope.status, 5)
        self.assertEqual(email_envelope.message, email)
        self.assertFalse(email_envelope.archive)
        self.assertEqual(email_envelope.message_id, sentinel.message_id)

    def test_from_dict_missing_email(self):
        with self.assertRaises(ValueError):
            EmailEnvelope.from_dict({})

    def test_from_dict_missing_email_data(self):
        with self.assertRaises(ValueError):
            EmailEnvelope.from_dict({'mesage_data': {'recipient': sentinel.recipient}})

    def test_from_dict_data(self):
        data = {'uid': sentinel.uid,
                'create_datetime': sentinel.create_datetime,
                'send_datetime': sentinel.send_datetime,
                'delivery_datetime': sentinel.delivery_datetime,
                'attempts': 1,
                'status': 5,
                'message_data': {'sender': None,
                                 'recipient': sentinel.recipient,
                                 'extra_headers': [],
                                 'subject_template': 'subject',
                                 'body_template': 'body',
                                 'body_template_html': None,
                                 'context': None},
                'archive': False,
                'message_id': sentinel.message_id}
        email_envelope = EmailEnvelope.from_dict(data, secretary_client=sentinel.secretary_client)
        self.assertEqual(email_envelope.uid, sentinel.uid)
        self.assertEqual(email_envelope.create_datetime, sentinel.create_datetime)
        self.assertEqual(email_envelope.send_datetime, sentinel.send_datetime)
        self.assertEqual(email_envelope.delivery_datetime, sentinel.delivery_datetime)
        self.assertEqual(email_envelope.attempts, 1)
        self.assertEqual(email_envelope.status, 5)
        self.assertFalse(email_envelope.archive)
        self.assertEqual(email_envelope.message_id, sentinel.message_id)
        self.assertIsInstance(email_envelope.message, Email)
        self.assertEqual(email_envelope.message.recipient, sentinel.recipient)
        self.assertEqual(email_envelope.message.secretary_client, sentinel.secretary_client)

    def test_to_dict(self):
        data = {
            'uid': sentinel.uid,
            'create_datetime': sentinel.create_datetime,
            'status': 5,
            'message_data': {
                'sender': None,
                'recipient': sentinel.recipient,
                'subject_template': 'subject',
                'body_template': 'body',
                'context': None,
            },
            'archive': False,
            'message_id': sentinel.message_id,
        }
        email_envelope = EmailEnvelope.from_dict(data)
        envelope_dict = email_envelope.to_dict()
        self.assertEqual(envelope_dict.pop('message'), email_envelope.message.to_dict())
        self.assertEqual(envelope_dict, {
            'uid': sentinel.uid,
            'create_datetime': sentinel.create_datetime,
            'send_datetime': None,
            'delivery_datetime': None,
            'attempts': 0,
            'status': 5,
            'archive': False,
            'message_id': sentinel.message_id,
        })
